# Standard Library Imports
import gettext
from typing import Any, Dict, Tuple

# Third Party Imports
from sqlalchemy.engine import Engine as Engine
from sqlalchemy.orm import scoped_session as scoped_session

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.db.base import BaseDatabase as BaseDatabase
from ramstk.models import RAMSTKSiteInfoRecord as RAMSTKSiteInfoRecord
from ramstk.models.commondb import RAMSTKRPN as RAMSTKRPN
from ramstk.models.commondb import RAMSTKCategory as RAMSTKCategory
from ramstk.models.commondb import RAMSTKCondition as RAMSTKCondition
from ramstk.models.commondb import RAMSTKFailureMode as RAMSTKFailureMode
from ramstk.models.commondb import RAMSTKGroup as RAMSTKGroup
from ramstk.models.commondb import RAMSTKHazards as RAMSTKHazards
from ramstk.models.commondb import RAMSTKLoadHistory as RAMSTKLoadHistory
from ramstk.models.commondb import RAMSTKManufacturer as RAMSTKManufacturer
from ramstk.models.commondb import RAMSTKMeasurement as RAMSTKMeasurement
from ramstk.models.commondb import RAMSTKMethod as RAMSTKMethod
from ramstk.models.commondb import RAMSTKModel as RAMSTKModel
from ramstk.models.commondb import RAMSTKStakeholders as RAMSTKStakeholders
from ramstk.models.commondb import RAMSTKStatus as RAMSTKStatus
from ramstk.models.commondb import RAMSTKSubCategory as RAMSTKSubCategory
from ramstk.models.commondb import RAMSTKType as RAMSTKType
from ramstk.models.commondb import RAMSTKUser as RAMSTKUser

_ = gettext.gettext
RAMSTK_CATEGORIES: Any
RAMSTK_CONDITIONS: Any
RAMSTK_GROUPS: Any
RAMSTK_FAILURE_MODES: Any
RAMSTK_HAZARDS: Any
RAMSTK_HISTORIES: Any
RAMSTK_MANUFACTURERS: Any
RAMSTK_MEASUREMENTS: Any
RAMSTK_METHODS: Any
RAMSTK_MODELS: Any
RAMSTK_RPNS: Any
RAMSTK_STAKEHOLDERS: Any
RAMSTK_STATUSES: Any
RAMSTK_SUBCATEGORIES: Any
RAMSTK_TYPES: Dict[int, Tuple[str, str, str]]

def _load_fmea_tables(session: scoped_session) -> None: ...
def _load_hazard_analysis_tables(session: scoped_session) -> None: ...
def _load_incident_report_tables(session: scoped_session) -> None: ...
def _load_miscellaneous_tables(session: scoped_session) -> None: ...
def _load_pof_tables(session: scoped_session) -> None: ...
def _load_requirements_analysis_tables(session: scoped_session) -> None: ...
def _load_site_info(session: scoped_session) -> None: ...
def do_add_administrator(session: scoped_session) -> None: ...
def do_create_common_db(engine: Engine, session: scoped_session) -> None: ...
def do_make_commondb_tables(engine: Engine) -> None: ...
def _do_load_action_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_hardware_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_incident_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_miscellaneous_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_pof_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_requirement_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_rpn_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_severity(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def _do_load_user_workgroups(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...
def do_load_variables(
    site_db: BaseDatabase, user_configuration: RAMSTKUserConfiguration
) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.failure_definition.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Failure Definition Panels."""

# Standard Library Imports
from typing import Any, Callable, Dict, List

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKTreePanel


class FailureDefinitionTreePanel(RAMSTKTreePanel):
    """Panel to display list of failure definitions."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_failure_definitions"
    _tag = "failure_definition"
    _title = _("Failure Definition List")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the failure definition panel."""
        super().__init__()

        # Initialize private dictionary class attributes.
        self.tvwTreeView.dic_row_loader: Dict[str, Callable] = {
            "failure_definition": self.__do_load_failure_definition,
        }

        # Initialize private list class attributes.

        # Initialize private scalar class attributes.
        self._on_edit_message: str = f"lvw_editing_{self._tag}"

        # Initialize public dictionary class attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "definition_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Definition ID"),
                "gint",
            ],
            "definition": [
                2,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Definition"),
                "gchararray",
            ],
        }

        # Initialize public list class attributes.

        # Initialize public scalar class attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _("Displays the list of failure definitions for the selected revision.")
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_module_switch, "lvwSwitchedPage")

    def _on_module_switch(self, module: str = "") -> None:
        """Respond to changes in selected Module View module (tab).

        :param module: the name of the module that was just selected.
        :return: None
        """
        _model, _row = self.tvwTreeView.selection.get_selected()

        if module == self._tag and _row is not None:
            _code = _model.get_value(_row, self.tvwTreeView.position["definition_id"])
            _title = _(f"Analyzing Failure Definition {_code}")

            pub.sendMessage("request_set_title", title=_title)

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Read attributes from newly selected RAMSTKTreeView() row.

        This method is called whenever a view's RAMSTKTreeView() row is
        activated/changed.

        :param selection: the Gtk.TreeSelection() for the newly selected row.
        :return: None
        """
        _attributes = super().on_row_change(selection)

        if _attributes:
            self._record_id = _attributes["definition_id"]

            pub.sendMessage("selected_failure_definition", attributes=_attributes)

    def __do_load_failure_definition(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load a failure definition into the RAMSTKTreeView().

        :param node: the treelib Node() with the definition data to load.
        :param row: the parent row of the definition to load.
        :return: _new_row; the row that was just populated with definition
            data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        _attributes = [_entity.revision_id, _entity.definition_id, _entity.definition]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading failure definition {0:s}.  "
                "This might indicate it was missing it's data package, some "
                "of the data in the package was missing, or some of the data "
                "was the wrong type.  Row data was: {1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row



# Standard Library Imports
from typing import Dict, List

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager as RAMSTKLogManager
from ramstk.views.gtk3.failure_definition import (
    FailureDefinitionListView as FailureDefinitionListView,
)
from ramstk.views.gtk3.widgets import RAMSTKBaseBook as RAMSTKBaseBook

class RAMSTKListBook(RAMSTKBaseBook):
    _dic_list_views: Dict[str, List[object]]
    _module: str
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _on_close(self) -> None: ...
    def _on_module_change(self, module: str = ...) -> None: ...



# Standard Library Imports
from typing import Any, Dict, List, Tuple

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.analyses import derating, stress
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

def do_check_overstress(
    overstress: Dict[str, List[float]], stress_type: str
) -> Tuple[bool, str]: ...

class RAMSTKDesignElectricRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    hardware_id: Any
    application_id: Any
    area: Any
    capacitance: Any
    configuration_id: Any
    construction_id: Any
    contact_form_id: Any
    contact_gauge: Any
    contact_rating_id: Any
    current_operating: Any
    current_rated: Any
    current_ratio: Any
    environment_active_id: Any
    environment_dormant_id: Any
    family_id: Any
    feature_size: Any
    frequency_operating: Any
    insert_id: Any
    insulation_id: Any
    manufacturing_id: Any
    matching_id: Any
    n_active_pins: Any
    n_circuit_planes: Any
    n_cycles: Any
    n_elements: Any
    n_hand_soldered: Any
    n_wave_soldered: Any
    operating_life: Any
    overstress: Any
    package_id: Any
    power_operating: Any
    power_rated: Any
    power_ratio: Any
    reason: Any
    resistance: Any
    specification_id: Any
    technology_id: Any
    temperature_active: Any
    temperature_case: Any
    temperature_dormant: Any
    temperature_hot_spot: Any
    temperature_junction: Any
    temperature_knee: Any
    temperature_rated_max: Any
    temperature_rated_min: Any
    temperature_rise: Any
    theta_jc: Any
    type_id: Any
    voltage_ac_operating: Any
    voltage_dc_operating: Any
    voltage_esd: Any
    voltage_rated: Any
    voltage_ratio: Any
    weight: Any
    years_in_production: Any
    def get_attributes(self) -> Dict[str, Any]: ...
    def do_calculate_current_ratio(self) -> None: ...
    def do_calculate_power_ratio(self) -> None: ...
    def do_calculate_voltage_ratio(self) -> None: ...
    def do_derating_analysis(self, stress_limits: List[float]) -> None: ...
    def do_stress_analysis(self, category_id: int) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.models.validation.table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Validation Table Model."""

# Standard Library Imports
from datetime import date
from typing import Any, Dict, Type

# Third Party Imports
import pandas as pd
from pubsub import pub

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseTable, RAMSTKValidationRecord


class RAMSTKValidationTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Validation data manager."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_validation_id"
    _db_tablename = "ramstk_validation"
    _select_msg = "selected_revision"
    _tag = "validation"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[Any, Any]) -> None:
        """Initialize a Validation data manager instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.
        self._dic_status: Dict[Any, float] = {}

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "validation_id",
        ]

        # Initialize private scalar attributes.
        self._record: Type[RAMSTKValidationRecord] = RAMSTKValidationRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "validation_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_calculate_plan, "request_calculate_plan")
        pub.subscribe(self._do_calculate_task, "request_calculate_validation_task")
        pub.subscribe(
            self._do_calculate_all_tasks, "request_calculate_validation_tasks"
        )

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Any]
    ) -> object:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.validation_id = self.last_id + 1
        _new_record.name = "New Validation Task"

        return _new_record

    def do_calculate_plan(self) -> None:
        """Calculate the planned burndown of the overall validation effort.

        This method will calculate three values for each scheduled end date
        in the validation plan, the lower bound, average, and upper bound
        for the tasks planned to still be open on each of the scheduled end
        dates in the plan.

        :return: _planned; the pandas DataFrame() containing the planned
            burndown hours for the entire validation effort.
        """
        _dic_plan = {}
        _dic_planned = {}  # type: ignore
        _time_ll = 0.0
        _time_mean = 0.0
        _time_ul = 0.0
        _start_date = date.today()

        for _node in self.tree.all_nodes()[1:]:
            # Calculate the three times if the mean task time is zero.
            if _node.data["validation"].time_mean <= 0.0:
                _node.data["validation"].calculate_task_time()

            # Keep a running total of the three times and the earliest task
            # start date.  The earliest start date will be assigned the
            # total number of hours in the validation program.
            _start_date = min(
                _start_date, pd.to_datetime(_node.data["validation"].date_start)
            )
            _time_ll += _node.data["validation"].time_ll
            _time_mean += _node.data["validation"].time_mean
            _time_ul += _node.data["validation"].time_ul

            # Calculate the sum of task hours for each, unique end date.
            _end_date = _node.data["validation"].date_end
            try:
                # Update the end date's times.
                _dic_planned[pd.to_datetime(_end_date)][0] += _node.data[
                    "validation"
                ].time_ll
                _dic_planned[pd.to_datetime(_end_date)][1] += _node.data[
                    "validation"
                ].time_mean
                _dic_planned[pd.to_datetime(_end_date)][2] += _node.data[
                    "validation"
                ].time_ul
            except KeyError:
                # Add the first time to the end date.
                _dic_planned[pd.to_datetime(_end_date)] = [
                    _node.data["validation"].time_ll,
                    _node.data["validation"].time_mean,
                    _node.data["validation"].time_ul,
                ]

        # Create a pandas DataFrame() of the task times sorted by date in
        # descending order.  The descending order is needed because the
        # ultimate DataFrame() will contain *remaining* total task hours for
        # the validation effort, not the total task hours planned to
        # complete on each day.
        # noinspection PyTypeChecker
        _planned = pd.DataFrame(
            _dic_planned.values(),
            index=_dic_planned.keys(),
            columns=["lower", "mean", "upper"],
        ).sort_index(ascending=False)

        # Calculate the total task time remaining on each planned end date
        # and then sort the DataFrame() by date in ascending order.
        _planned = _planned.cumsum() - _planned
        _planned.loc[_start_date] = [_time_ll, _time_mean, _time_ul]
        _planned = _planned.sort_index()

        _dic_plan["plan"] = _planned
        _dic_plan["assessed"] = self._do_select_assessment_targets()

        pub.sendMessage(
            "succeed_calculate_verification_plan",
            attributes=_dic_plan,
        )

    def _do_calculate_all_tasks(self) -> None:
        """Calculate mean, standard error, and bounds on all task's time/cost.

        These values are calculated assuming a beta distribution (typical
        project management assumption).  This method also calculates the
        remaining average time and cost of the overall validation plan.

        :return: None
        :rtype: None
        """
        _program_cost_remaining = 0.0
        _program_time_remaining = 0.0

        for _node in self.tree.all_nodes()[1:]:
            self._do_calculate_task(_node.identifier)

            _program_cost_remaining += _node.data["validation"].cost_average * (
                1.0 - _node.data["validation"].status / 100.0
            )
            _program_time_remaining += _node.data["validation"].time_average * (
                1.0 - _node.data["validation"].status / 100.0
            )

        pub.sendMessage(
            "succeed_calculate_all_validation_tasks",
            cost_remaining=_program_cost_remaining,
            time_remaining=_program_time_remaining,
        )

    def _do_calculate_task(self, node_id: int) -> None:
        """Calculate mean, standard error, and bounds on task time and cost.

        These values are calculated assuming a beta distribution (typical
        project management assumption).

        :param node_id: the ID of the node (task) to calculate.
        :return: None
        :rtype: None
        """
        _node = self.tree.get_node(node_id)

        _node.data["validation"].calculate_task_time()
        _node.data["validation"].calculate_task_cost()

        self.do_set_attributes_all(attributes=_node.data["validation"].get_attributes())

    def _do_select_assessment_targets(self) -> pd.DataFrame:
        """Select the targets for all tasks of Reliability Assessment type.

        :return: _assessed; a pandas DataFrame() containing the assessment
            dates as the index and associated targets.
        """
        _dic_assessed = {}
        for _node in self.tree.all_nodes()[1:]:
            if _node.data["validation"].task_type == 5:
                _dic_assessed[pd.to_datetime(_node.data["validation"].date_end)] = [
                    _node.data["validation"].acceptable_minimum,
                    _node.data["validation"].acceptable_mean,
                    _node.data["validation"].acceptable_maximum,
                ]

        # noinspection PyTypeChecker
        return pd.DataFrame(
            _dic_assessed.values(),
            index=_dic_assessed.keys(),
            columns=["lower", "mean", "upper"],
        ).sort_index()



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.criticality.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""FMEA Criticality Analysis Module."""

# RAMSTK Package Imports
from ramstk.exceptions import OutOfRangeError


def calculate_rpn(sod):
    """Calculate the Risk Priority Number (RPN).

        RPN = S * O * D

        >>> SOD = {'rpn_severity': 5, 'rpn_occurrence': 8, 'rpn_detection': 7}
        >>> criticality.calculate_rpn(SOD)
        280

        >>> SOD['rpn_severity'] = 0
        >>> criticality.calculate_rpn(SOD)
        Traceback (most recent call last):
        ...
        ramstk.exceptions.OutOfRangeError: RPN severity is outside the range [1, 10].   # noqa


    :param sod: a dict containing the S, O, and D values from which to
        calculate the RPN.
    :return: _rpn; the calculated risk priority number (RPN).
    :rtype: int
    :raise: OutOfRangeError if one of the inputs falls outside the range
        [1, 10].
    """
    if not 0 < sod["rpn_severity"] < 11:
        raise OutOfRangeError(("RPN severity is outside the range [1, 10]."))
    if not 0 < sod["rpn_occurrence"] < 11:
        raise OutOfRangeError(("RPN occurrence is outside the range [1, 10]."))
    if not 0 < sod["rpn_detection"] < 11:
        raise OutOfRangeError(("RPN detection is outside the range [1, 10]."))

    _rpn = (
        int(sod["rpn_severity"])
        * int(sod["rpn_occurrence"])
        * int(sod["rpn_detection"])
    )

    return _rpn


def calculate_mode_hazard_rate(item_hr, mode_ratio):
    """Calculate the failure mode hazard rate.

        >>> item_hr=0.000617
        >>> mode_ratio=0.23
        >>> calculate_mode_hazard_rate(item_hr, mode_ratio)
        0.00014191

    :param item_hr: the hazard rate of the (hardware) item the mode is
        associated with.
    :param mode_ratio:
    :return: _mode_hazard_rate; the hazard rate of the failure mode.
    :rtype: float
    :raise: OutOfRangeError if passed a negative item hazard rate or a mode
        ratio outside [0.0, 1.0].
    """
    if item_hr < 0.0:
        raise OutOfRangeError(
            (
                "calculate_mode_hazard_rate() was passed a "
                "negative value for the item hazard rate."
            )
        )
    if not 0.0 <= mode_ratio <= 1.0:
        raise OutOfRangeError(
            (
                "calculate_mode_hazard_rate() was passed a "
                "failure mode ratio outside the range of "
                "[0.0, 1.0]."
            )
        )

    return item_hr * mode_ratio


def calculate_mode_criticality(mode_hr, mode_op_time, eff_prob):
    """Calculate the MIL-HDBK-1629A, Task 102 criticality.

        >>> mode_hr=0.00021595
        >>> mode_op_time=4.15
        >>> eff_prob=1.0
        >>> calculate_mode_criticality(mode_hr, mode_op_time, eff_prob)

    :param mode_hr: the hazard rate of the failure mode.
    :param mode_op_time: the operating time of the failure mode (i.e.,
        the time at risk).
    :param eff_prob: the probability the end effect is the one observed.
    :return: _criticality; the calculated Task 102 criticality.
    :rtype: float
    :raise: OutOfRangeError if passed a negative mode operating time or an
        effect probability outside [0.0, 1.0].
    """
    if mode_op_time < 0.0:
        raise OutOfRangeError(
            (
                "calculate_mode_criticality() was passed a "
                "negative value for failure mode operating "
                "time."
            )
        )
    if not 0.0 <= eff_prob <= 1.0:
        raise OutOfRangeError(
            (
                "calculate_mode_criticality() was passed a "
                "failure effect probability outside the range "
                "of [0.0, 1.0]."
            )
        )

    return mode_hr * mode_op_time * eff_prob



# Standard Library Imports
from typing import Any, Callable, Dict, List, Tuple

# RAMSTK Package Imports
from ramstk.utilities import deprecated as deprecated
from ramstk.utilities import string_to_boolean as string_to_boolean
from ramstk.views.gtk3 import Gdk as Gdk
from ramstk.views.gtk3 import GdkPixbuf as GdkPixbuf
from ramstk.views.gtk3 import GObject as GObject
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import Pango as Pango

# RAMSTK Local Imports
from .label import RAMSTKLabel as RAMSTKLabel
from .widget import RAMSTKWidget as RAMSTKWidget

def do_make_column(
    cells: List[object], **kwargs: Dict[str, Any]
) -> Gtk.TreeViewColumn: ...
def do_set_cell_properties(cell: object, properties: Dict[str, Any]) -> None: ...

class RAMSTKTreeView(Gtk.TreeView, RAMSTKWidget):
    _has_pixbuf: bool
    dic_row_loader: Dict[str, Callable]
    cellprops: Dict[str, Any]
    datatypes: Dict[str, str]
    editable: Dict[str, bool]
    headings: Dict[str, str]
    position: Dict[str, int]
    visible: Dict[str, bool]
    widgets: Dict[str, object]
    selection: Gtk.TreeSelection
    filt_model: Gtk.TreeModelFilter
    unfilt_model: Gtk.TreeModel
    def __init__(self) -> None: ...
    def do_edit_cell(
        self, cell: Gtk.CellRenderer, path: str, new_text: Any, position: int
    ) -> Any: ...
    def do_expand_tree(self) -> None: ...
    def do_get_row_by_value(self, search_col: int, value: Any) -> Gtk.TreeIter: ...
    def do_insert_row(self, data: Dict[str, Any], prow: Gtk.TreeIter = ...) -> None: ...
    def do_load_combo_cell(self, index: int, items: List[str]) -> None: ...
    def do_make_columns(self) -> None: ...
    def do_make_model(self) -> None: ...
    def do_parse_format(self, fmt_file: str) -> None: ...
    def do_set_editable_columns(self, method: object) -> None: ...
    def do_set_visible_columns(self) -> None: ...
    def get_aggregate_attributes(self, entity: object) -> List[Any]: ...
    def get_cell_model(self, column: int, clear: bool = ...) -> Gtk.TreeModel: ...
    def get_simple_attributes(self, entity: object) -> List[Any]: ...
    @staticmethod
    def _do_format_cell(
        __column: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        row: Gtk.TreeIter,
        data: Tuple[Any],
    ) -> None: ...
    def _do_set_column_properties(
        self, key: str, column: Gtk.TreeViewColumn
    ) -> None: ...
    @staticmethod
    def _resize_wrap(
        column: Gtk.TreeViewColumn, __param, cell: Gtk.CellRenderer
    ) -> None: ...

class CellRendererML(Gtk.CellRendererText):
    textedit_window: Any
    selection: Any
    treestore: Any
    treeiter: Any
    textedit: Any
    textbuffer: Any
    def __init__(self) -> None: ...
    def do_get_size(self, widget, cell_area): ...
    def do_start_editing(
        self, __event, treeview, path, __background_area, cell_area, __flags
    ) -> None: ...
    def _keyhandler(self, __widget, event) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.design_electric.components.resistor.py is part of the
#       RAMSTK Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Resistor Input Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.widgets import RAMSTKComboBox, RAMSTKEntry, RAMSTKFixedPanel


class ResistorDesignElectricInputPanel(RAMSTKFixedPanel):
    """Display Resistor assessment input attribute data.

    The Resistor assessment input view displays all the assessment inputs for
    the selected resistor.  This includes, currently, inputs for
    MIL-HDBK-217FN2.  The attributes of a Resistor assessment input view are:

    :cvar dict _dic_specifications: dictionary of resistor MIL-SPECs.  Key is
        resistor subcategory ID; values are lists of specifications.
    :cvar dict _dic_styles: dictionary of resistor styles defined in the
        MIL-SPECs.  Key is resistor subcategory ID; values are lists of styles.

    :ivar cmbSpecification: select and display the governing specification of
        the resistor.
    :ivar cmbType: select and display the type of thermistor.
    :ivar cmbConstruction: select and display the method of construction of the
        resistor.
    :ivar txtResistance: enter and display the resistance of the resistor.
    :ivar txtNElements: enter and display the number of active resistors in a
        resistor network or the number of potentiometers taps.
    """

    # Define private dict class attributes.
    _dic_quality: Dict[int, List[List[str]]] = {
        1: [["S"], ["R"], ["P"], ["M"], ["MIL-R-11"], [_("Lower")]],
        2: [["S"], ["R"], ["P"], ["M"], ["MIL-R-10509"], ["MIL-R-22684"], [_("Lower")]],
        3: [["MIL-SPEC"], [_("Lower")]],
        4: [["MIL-SPEC"], [_("Lower")]],
        5: [["S"], ["R"], ["P"], ["M"], ["MIL-R-93"], [_("Lower")]],
        6: [["S"], ["R"], ["P"], ["M"], ["MIL-R-26"], [_("Lower")]],
        7: [["S"], ["R"], ["P"], ["M"], ["MIL-R-18546"], [_("Lower")]],
        8: [["MIL-SPEC"], [_("Lower")]],
        9: [["S"], ["R"], ["P"], ["M"], ["MIL-R-27208"], [_("Lower")]],
        10: [["MIL-SPEC"], [_("Lower")]],
        11: [["MIL-SPEC"], [_("Lower")]],
        12: [["MIL-SPEC"], [_("Lower")]],
        13: [["S"], ["R"], ["P"], ["M"], ["MIL-R-22097"], [_("Lower")]],
        14: [["MIL-SPEC"], [_("Lower")]],
        15: [["MIL-SPEC"], [_("Lower")]],
    }
    # Key is subcategory ID; index is specification ID.
    _dic_specifications: Dict[int, List[List[str]]] = {
        2: [["MIL-R-10509"], ["MIL-R-22684"], ["MIL-R-39017"], ["MIL-R-55182"]],
        6: [["MIL-R-26"], ["MIL-R-39007"]],
        7: [["MIL-R-18546"], ["MIL-R-39009"]],
        15: [["MIL-R-23285"], ["MIL-R-39023"]],
    }
    # Key is subcategory ID, index is type ID.
    _dic_types: Dict[int, List[List[str]]] = {
        1: [["RCR"], ["RC"]],
        2: [["RLR"], ["RL"], ["RNR"], ["RN"]],
        5: [["RBR"], ["RB"]],
        6: [["RWR"], ["RW"]],
        7: [["RER"], ["RE"]],
        9: [["RTR"], ["RT"]],
        11: [["RA"], ["RK"]],
        13: [["RJR"], ["RJ"]],
        15: [["RO"], ["RVC"]],
    }
    # First key is subcategory ID; second key is specification ID.
    # Index is style ID.
    _dic_styles: Dict[int, Dict[int, List[List[str]]]] = {
        6: {
            1: [
                ["RWR 71"],
                ["RWR 74"],
                ["RWR 78"],
                ["RWR 80"],
                ["RWR 81"],
                ["RWR 82"],
                ["RWR 84"],
                ["RWR 89"],
            ],
            2: [
                ["RW 10"],
                ["RW 11"],
                ["RW 12"],
                ["RW 13"],
                ["RW 14"],
                ["RW 15"],
                ["RW 16"],
                ["RW 20"],
                ["RW 21 "],
                ["RW 22"],
                ["RW 23"],
                ["RW 24"],
                ["RW 29"],
                ["RW 30"],
                ["RW 31"],
                ["RW 32"],
                ["RW 33"],
                ["RW 34"],
                ["RW 35"],
                ["RW 36"],
                ["RW 37"],
                ["RW 38"],
                ["RW 39"],
                ["RW 47"],
                ["RW 55"],
                ["RW 56"],
                ["RW 67"],
                ["RW 68"],
                ["RW 69"],
                ["RW 70"],
                ["RW 74"],
                ["RW 78"],
                ["RW 79"],
                ["RW 80"],
                ["RW 81"],
            ],
        },
        7: {
            1: [
                ["RE 60/RER 60"],
                ["RE 65/RER 65"],
                ["RE 70/RER 70"],
                ["RE 75/RER 75"],
                ["RE 77"],
                ["RE 80"],
            ],
            2: [
                ["RE 60/RER40"],
                ["RE 65/RER 45"],
                ["RE 70/ RER 50"],
                ["RE 75/RER 55"],
                ["RE 77"],
                ["RE 80"],
            ],
        },
    }
    # Key is subcategory ID; index is construction ID.
    _dic_construction: Dict[int, List[List[str]]] = {
        10: [
            ["RR0900A2A9J103"],
            ["RR0900A3A9J103"],
            ["RR0900A4A9J103"],
            ["RR0900A5A9J103"],
        ],
        12: [[_("Enclosed")], [_("Unenclosed")]],
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field: str = "hardware_id"
    _select_msg: str = "selected_hardware"
    _tag: str = "design_electric"
    _title: str = _("Resistor Design Inputs")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Resistor assessment input view."""
        super().__init__()

        # Initialize widgets.
        self.cmbConstruction: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbQuality: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbSpecification: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbStyle: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbType: RAMSTKComboBox = RAMSTKComboBox()
        self.txtNElements: RAMSTKEntry = RAMSTKEntry()
        self.txtResistance: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._quality_id: int = 0

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "quality_id": [
                32,
                self.cmbQuality,
                "changed",
                super().on_changed_combo,
                "wvw_editing_reliability",
                0,
                {
                    "tooltip": _("The quality level of the resistor."),
                },
                _("Quality Level:"),
                "gint",
            ],
            "resistance": [
                35,
                self.txtResistance,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The resistance (in \u03A9) of the resistor."),
                },
                _("Resistance (\u03A9):"),
                "gfloat",
            ],
            "specification_id": [
                36,
                self.cmbSpecification,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The governing specification for the resistor."),
                },
                _("Specification:"),
                "gint",
            ],
            "type_id": [
                48,
                self.cmbType,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The type of thermistor."),
                },
                _("Type:"),
                "gint",
            ],
            "family_id": [
                15,
                self.cmbStyle,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The style of resistor."),
                },
                _("Style:"),
                "gint",
            ],
            "construction_id": [
                6,
                self.cmbConstruction,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The method of construction of the resistor."),
                },
                _("Construction:"),
                "gint",
            ],
            "n_elements": [
                2,
                self.txtNElements,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _(
                        "The number of active resistors in a resistor network or the "
                        "number of potentiometer taps."
                    ),
                },
                _("Number of Elements:"),
                "gint",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.category_id: int = 0
        self.subcategory_id: int = 0

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_load_comboboxes,
            "changed_subcategory",
        )
        pub.subscribe(
            self._do_set_reliability_attributes,
            "succeed_get_reliability_attributes",
        )

    def do_load_comboboxes(self, subcategory_id: int) -> None:
        """Load the Resistor RKTComboBox()s.

        :param subcategory_id: the subcategory ID of the selected capacitor.
            This is unused in this method but required because this method is a
            PyPubSub listener.
        :return: None
        :rtype: None
        """
        self.subcategory_id = subcategory_id

        self.__do_load_quality_combo()
        self.__do_load_specification_combo()
        self.__do_load_type_combo()
        self.__do_load_style_combo()
        self.__do_load_construction_combo()

    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self._hazard_rate_method_id = attributes["hazard_rate_method_id"]
            self._quality_id = attributes["quality_id"]

            self.cmbQuality.set_sensitive(True)
            self.cmbQuality.do_update(
                self._quality_id,
                signal="changed",
            )

            pub.sendMessage(
                f"request_get_{self._tag}_attributes",
                node_id=self._record_id,
            )

    def _do_set_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set widget sensitivity as needed for the selected resistor.

        :param attributes: the dict of design electric attributes.
        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2:
            self.txtResistance.set_sensitive(True)
            self.txtResistance.do_update(
                str(self.fmt.format(attributes["resistance"] or 0.0)),
                signal="changed",
            )
        else:
            self.txtResistance.set_sensitive(False)

        self.__do_set_construction_combo_sensitive(attributes)
        self.__do_set_elements_entry_sensitive(attributes)
        self.__do_set_specification_combo_sensitive(attributes)
        self.__do_set_style_combo_sensitive(attributes)
        self.__do_set_type_combo_sensitive(attributes)

    def __do_load_construction_combo(self) -> None:
        """Load the Resistor construction RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        try:
            _data = self._dic_construction[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbConstruction.do_load_combo(
            _data,
            signal="changed",
        )

    def __do_load_quality_combo(self) -> None:
        """Load the Resistor quality RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        try:
            if self._hazard_rate_method_id == 1:
                _quality: List[List[str]] = [
                    ["S"],
                    ["R"],
                    ["P"],
                    ["M"],
                    ["MIL-SPEC"],
                    [_("Lower")],
                ]
            else:
                _quality = self._dic_quality[self.subcategory_id]
        except KeyError:
            _quality = [[""]]
        self.cmbQuality.do_load_combo(
            entries=_quality,
            signal="changed",
        )

    def __do_load_specification_combo(self) -> None:
        """Load the Resistor specification RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        try:
            _data = self._dic_specifications[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbSpecification.do_load_combo(
            _data,
            signal="changed",
        )

    def __do_load_style_combo(self) -> None:
        """Load the Resistor style RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        _specification_id = int(self.cmbSpecification.get_active())
        try:
            _styles: List[List[str]] = self._dic_styles[self.subcategory_id][
                _specification_id
            ]
        except (KeyError, IndexError):
            _styles = [[""]]
        self.cmbStyle.do_load_combo(
            entries=_styles,
            signal="changed",
        )

    def __do_load_type_combo(self) -> None:
        """Load the Resistor (thermistor) type RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        try:
            if self._hazard_rate_method_id == 1:
                _types: List[Any] = self._dic_types[self.subcategory_id]
            else:
                _types = [[_("Bead")], [_("Disk")], [_("Rod")]]
        except KeyError:
            _types = []
        self.cmbType.do_load_combo(
            entries=_types,
            signal="changed",
        )

    def __do_set_construction_combo_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the Resistor construction RAMSTKComboBox() sensitive or not.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2 and self.subcategory_id in [10, 12]:
            self.cmbConstruction.set_sensitive(True)
            self.cmbConstruction.do_update(
                attributes["construction_id"],
                signal="changed",
            )
        else:
            self.cmbConstruction.set_sensitive(False)

    def __do_set_elements_entry_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set Resistor number of elements RAMSTKEntry() sensitive or not.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2 and self.subcategory_id in [
            4,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
        ]:
            self.txtNElements.set_sensitive(True)
            self.txtNElements.do_update(
                str(attributes["n_elements"]),
                signal="changed",
            )
        else:
            self.txtNElements.set_sensitive(False)

    def __do_set_specification_combo_sensitive(
        self, attributes: Dict[str, Any]
    ) -> None:
        """Set the Resistor specification RAMSTKComboBox() sensitive or not.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2 and self.subcategory_id in [2, 6, 7, 15]:
            self.cmbSpecification.set_sensitive(True)
            self.cmbSpecification.do_update(
                attributes["specification_id"],
                signal="changed",
            )
        else:
            self.cmbSpecification.set_sensitive(False)

    def __do_set_style_combo_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the Resistor style RAMSTKComboBox() sensitive or not.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2 and self.subcategory_id in [6, 7]:
            self.cmbStyle.set_sensitive(True)
            self.cmbStyle.do_update(
                attributes["family_id"],
                signal="changed",
            )
        else:
            self.cmbStyle.set_sensitive(False)

    def __do_set_type_combo_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the Resistor type RAMSTKComboBox() sensitive or not.

        :return: None
        :rtype: None
        """
        self.cmbType.set_sensitive(False)

        if self._hazard_rate_method_id == 1 and self.subcategory_id in [
            1,
            2,
            5,
            6,
            7,
            9,
            11,
            13,
            15,
        ]:
            self.cmbType.set_sensitive(True)
            self.cmbType.do_update(
                attributes["type_id"],
                signal="changed",
            )
        elif self._hazard_rate_method_id == 2 and self.subcategory_id == 8:
            self.cmbType.set_sensitive(True)



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.similar_item.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Similar Item Panels."""

# Standard Library Imports
from typing import Any, Dict, List, Union

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKComboBox, RAMSTKFixedPanel, RAMSTKTreePanel


class SimilarItemMethodPanel(RAMSTKFixedPanel):
    """Panel to display Similar Item analysis methods."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "selected_similar_item"
    _tag = "similar_item"
    _title = _("Similar Item Method")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self):
        """Initialize an instance of the Similar Item methods panel."""
        super().__init__()

        # Initialize widgets.
        self.cmbSimilarItemMethod: RAMSTKComboBox = RAMSTKComboBox()

        # Initialize private dictionary instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._method_id: int = 0
        self._on_edit_message = f"wvw_editing_{self._tag}"

        # Initialize public dictionary instance attributes.
        self.dic_attribute_widget_map = {
            "similar_item_method_id": [
                29,
                self.cmbSimilarItemMethod,
                "changed",
                super().on_changed_combo,
                self._on_edit_message,
                0,
                {
                    "tooltip": _("Select the similar item analysis method."),
                },
                _("Select Similar Item Method "),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.cmbSimilarItemMethod.connect("changed", self._on_method_changed)

        # Subscribe to PyPubSub messages.

    def do_load_comboboxes(self) -> None:
        """Load Similar Item analysis RAMSTKComboBox()s.

        :return: None
        :rtype: None
        """
        # Load the method combobox.
        self.cmbSimilarItemMethod.do_load_combo(
            [[_("Topic 633"), 0], [_("User-Defined"), 1]], signal="changed"
        )

    @staticmethod
    def _on_method_changed(combo: RAMSTKComboBox) -> None:
        """Let others know when similar item method combo changes.

        :param combo: the similar item calculation method RAMSTKComboBox().
        :return: None
        :rtype: None
        """
        pub.sendMessage(
            "succeed_change_similar_item_method", method_id=combo.get_active()
        )


class SimilarItemTreePanel(RAMSTKTreePanel):
    """Panel to display Similar Item analysis worksheet."""

    # Define private dict class attributes.
    _dic_quality: Dict[int, str] = {
        0: "",
        1: "Space",
        2: "Full Military",
        3: "Ruggedized",
        4: "Commercial",
    }
    _dic_environment: Dict[int, str] = {
        0: "",
        1: "Ground, Benign",
        2: "Ground,Mobile",
        3: "Naval, Sheltered",
        4: "Airborne, Inhabited, Cargo",
        5: "Airborne, Rotary Wing",
        6: "Space, Flight",
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_similar_items"
    _tag = "similar_item"
    _title = _("Similar Item Analysis")

    # Define public dictionary class attributes.

    # Define public dictionary list attributes.

    # Define public dictionary scalar attributes.

    # Define private dictionary class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Similar Item analysis worksheet."""
        super().__init__()

        # Initialize private dict instance attributes.
        self._dic_hardware_attrs: Dict[int, List[Union[bool, float, int, str]]] = {}
        self._dic_reliability_attrs: Dict[int, List[Union[bool, float, int, str]]] = {}
        self.tvwTreeView.dic_row_loader = {
            "similar_item": self.__do_load_similar_item,
        }
        self._dic_visible_mask: Dict[int, Dict[str, bool]] = {
            0: {
                "revision_id": False,
                "hardware_id": False,
                "name": True,
                "hazard_rate_active": True,
                "quality_from_id": False,
                "quality_to_id": False,
                "environment_from_id": False,
                "environment_to_id": False,
                "temperature_from": False,
                "temperature_to": False,
                "change_description_1": False,
                "change_factor_1": False,
                "change_description_2": False,
                "change_factor_2": False,
                "change_description_3": False,
                "change_factor_3": False,
                "change_description_4": False,
                "change_factor_4": False,
                "change_description_5": False,
                "change_factor_5": False,
                "change_description_6": False,
                "change_factor_6": False,
                "change_description_7": False,
                "change_factor_7": False,
                "change_description_8": False,
                "change_factor_8": False,
                "change_description_9": False,
                "change_factor_9": False,
                "change_description_10": False,
                "change_factor_10": False,
                "function_1": False,
                "function_2": False,
                "function_3": False,
                "function_4": False,
                "function_5": False,
                "result_1": True,
                "result_2": False,
                "result_3": False,
                "result_4": False,
                "result_5": False,
                "user_blob_1": False,
                "user_blob_2": False,
                "user_blob_3": False,
                "user_blob_4": False,
                "user_blob_5": False,
                "user_float_1": False,
                "user_float_2": False,
                "user_float_3": False,
                "user_float_4": False,
                "user_float_5": False,
                "user_int_1": False,
                "user_int_2": False,
                "user_int_3": False,
                "user_int_4": False,
                "user_int_5": False,
            },
            1: {
                "revision_id": False,
                "hardware_id": False,
                "name": True,
                "hazard_rate_active": True,
                "quality_from_id": True,
                "quality_to_id": True,
                "environment_from_id": True,
                "environment_to_id": True,
                "temperature_from": True,
                "temperature_to": True,
                "change_description_1": False,
                "change_factor_1": False,
                "change_description_2": False,
                "change_factor_2": False,
                "change_description_3": False,
                "change_factor_3": False,
                "change_description_4": False,
                "change_factor_4": False,
                "change_description_5": False,
                "change_factor_5": False,
                "change_description_6": False,
                "change_factor_6": False,
                "change_description_7": False,
                "change_factor_7": False,
                "change_description_8": False,
                "change_factor_8": False,
                "change_description_9": False,
                "change_factor_9": False,
                "change_description_10": False,
                "change_factor_10": False,
                "function_1": False,
                "function_2": False,
                "function_3": False,
                "function_4": False,
                "function_5": False,
                "result_1": True,
                "result_2": False,
                "result_3": False,
                "result_4": False,
                "result_5": False,
                "user_blob_1": False,
                "user_blob_2": False,
                "user_blob_3": False,
                "user_blob_4": False,
                "user_blob_5": False,
                "user_float_1": False,
                "user_float_2": False,
                "user_float_3": False,
                "user_float_4": False,
                "user_float_5": False,
                "user_int_1": False,
                "user_int_2": False,
                "user_int_3": False,
                "user_int_4": False,
                "user_int_5": False,
            },
            2: {
                "revision_id": False,
                "hardware_id": False,
                "name": True,
                "hazard_rate_active": True,
                "quality_from_id": False,
                "quality_to_id": False,
                "environment_from_id": False,
                "environment_to_id": False,
                "temperature_from": False,
                "temperature_to": False,
                "change_description_1": True,
                "change_factor_1": True,
                "change_description_2": True,
                "change_factor_2": True,
                "change_description_3": True,
                "change_factor_3": True,
                "change_description_4": True,
                "change_factor_4": True,
                "change_description_5": True,
                "change_factor_5": True,
                "change_description_6": True,
                "change_factor_6": True,
                "change_description_7": True,
                "change_factor_7": True,
                "change_description_8": True,
                "change_factor_8": True,
                "change_description_9": True,
                "change_factor_9": True,
                "change_description_10": True,
                "change_factor_10": True,
                "function_1": False,
                "function_2": False,
                "function_3": False,
                "function_4": False,
                "function_5": False,
                "result_1": True,
                "result_2": True,
                "result_3": True,
                "result_4": True,
                "result_5": True,
                "user_blob_1": True,
                "user_blob_2": True,
                "user_blob_3": True,
                "user_blob_4": True,
                "user_blob_5": True,
                "user_float_1": True,
                "user_float_2": True,
                "user_float_3": True,
                "user_float_4": True,
                "user_float_5": True,
                "user_int_1": True,
                "user_int_2": True,
                "user_int_3": True,
                "user_int_4": True,
                "user_int_5": True,
            },
        }

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._method_id: int = 0
        self._on_edit_message: str = f"mvw_editing_{self._tag}"

        # Initialize public dictionary instance attributes.
        self.dic_attribute_widget_map = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "hardware_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Hardware ID"),
                "gint",
            ],
            "name": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Assembly"),
                "gchararray",
            ],
            "hazard_rate_active": [
                3,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Current Hazard Rate"),
                "gfloat",
            ],
            "quality_from_id": [
                4,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("From Quality"),
                "gchararray",
            ],
            "quality_to_id": [
                5,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("To Quality"),
                "gchararray",
            ],
            "environment_from_id": [
                6,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("From Environment"),
                "gchararray",
            ],
            "environment_to_id": [
                7,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("To Environment"),
                "gchararray",
            ],
            "temperature_from": [
                8,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                25.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("From Temperature"),
                "gfloat",
            ],
            "temperature_to": [
                9,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                25.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("To Temperature"),
                "gfloat",
            ],
            "change_description_1": [
                10,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 1"),
                "gchararray",
            ],
            "change_factor_1": [
                11,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Change Factor 1"),
                "gfloat",
            ],
            "change_description_2": [
                12,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 2"),
                "gchararray",
            ],
            "change_factor_2": [
                13,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 2"),
                "gfloat",
            ],
            "change_description_3": [
                14,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 3"),
                "gchararray",
            ],
            "change_factor_3": [
                15,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 3"),
                "gfloat",
            ],
            "change_description_4": [
                16,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 4"),
                "gchararray",
            ],
            "change_factor_4": [
                17,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 4"),
                "gfloat",
            ],
            "change_description_5": [
                18,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 5"),
                "gchararray",
            ],
            "change_factor_5": [
                19,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 5"),
                "gfloat",
            ],
            "change_description_6": [
                20,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 6"),
                "gchararray",
            ],
            "change_factor_6": [
                21,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 6"),
                "gfloat",
            ],
            "change_description_7": [
                22,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 7"),
                "gchararray",
            ],
            "change_factor_7": [
                23,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 7"),
                "gfloat",
            ],
            "change_description_8": [
                24,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 8"),
                "gchararray",
            ],
            "change_factor_8": [
                25,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 8"),
                "gfloat",
            ],
            "change_description_9": [
                26,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 9"),
                "gchararray",
            ],
            "change_factor_9": [
                27,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 9"),
                "gfloat",
            ],
            "change_description_10": [
                28,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Description 10"),
                "gchararray",
            ],
            "change_factor_10": [
                29,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Change Factor 10"),
                "gfloat",
            ],
            "function_1": [
                30,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Function 1"),
                "gchararray",
            ],
            "function_2": [
                31,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Function 2"),
                "gchararray",
            ],
            "function_3": [
                32,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Function 3"),
                "gchararray",
            ],
            "function_4": [
                33,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Function 4"),
                "gchararray",
            ],
            "function_5": [
                34,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Function 5"),
                "gchararray",
            ],
            "result_1": [
                35,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Result 1"),
                "gfloat",
            ],
            "result_2": [
                36,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Result 2"),
                "gfloat",
            ],
            "result_3": [
                37,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Result 3"),
                "gfloat",
            ],
            "result_4": [
                38,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Result 4"),
                "gfloat",
            ],
            "result_5": [
                39,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Result 5"),
                "gfloat",
            ],
            "user_blob_1": [
                40,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Text 1"),
                "gchararray",
            ],
            "user_blob_2": [
                41,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Text 2"),
                "gchararray",
            ],
            "user_blob_3": [
                42,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Text 3"),
                "gchararray",
            ],
            "user_blob_4": [
                43,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Text 4"),
                "gchararray",
            ],
            "user_blob_5": [
                44,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Text 5"),
                "gchararray",
            ],
            "user_float_1": [
                45,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Float 1"),
                "gfloat",
            ],
            "user_float_2": [
                46,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Float 2"),
                "gfloat",
            ],
            "user_float_3": [
                47,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Float 3"),
                "gfloat",
            ],
            "user_float_4": [
                48,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Float 4"),
                "gfloat",
            ],
            "user_float_5": [
                49,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Float 5"),
                "gfloat",
            ],
            "user_int_1": [
                50,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Integer 1"),
                "gint",
            ],
            "user_int_2": [
                51,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Integer 2"),
                "gint",
            ],
            "user_int_3": [
                52,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Integer 3"),
                "gint",
            ],
            "user_int_4": [
                53,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Integer 4"),
                "gint",
            ],
            "user_int_5": [
                54,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("User Integer 5"),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _(
                "Displays the Similar Item Analysis for the currently selected "
                "Hardware item."
            )
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(super().do_load_panel, "succeed_calculate_similar_item")
        pub.subscribe(self._do_set_hardware_attributes, "succeed_retrieve_hardwares")
        pub.subscribe(
            self._do_set_reliability_attributes, "succeed_retrieve_reliabilitys"
        )
        pub.subscribe(self._on_method_changed, "succeed_change_similar_item_method")

    def do_load_comboboxes(self) -> None:
        """Load Similar Item analysis RAMSTKComboBox()s.

        :return: None
        :rtype: None
        """
        # Load the quality from and quality to Gtk.CellRendererCombo().
        for _idx in [4, 5]:
            _model = self.tvwTreeView.get_cell_model(_idx, True)
            for _quality in self._dic_quality.values():
                _model.append([_quality])

        # Load the environment from and environment to Gtk.CellRendererCombo().
        for _idx in [6, 7]:
            _model = self.tvwTreeView.get_cell_model(_idx, True)
            for _environment in self._dic_environment.values():
                _model.append([_environment])

    def do_refresh_functions(self, row: Gtk.TreeIter, function: List[str]) -> None:
        """Refresh the Similar Item functions in the RAMSTKTreeView().

        :param row: the row in the Similar Item RAMSTKTreeView() whose
            functions need to be updated.  This is require to allow a recursive
            calling function to load the same function in all rows.
        :param function: the list of user-defined Similar Item functions.
        :return: None
        """
        _model = self.tvwTreeView.get_model()

        _model.set_value(row, self.tvwTreeView.position["function_1"], function[0])
        _model.set_value(row, self.tvwTreeView.position["function_2"], function[1])
        _model.set_value(row, self.tvwTreeView.position["function_3"], function[2])
        _model.set_value(row, self.tvwTreeView.position["function_4"], function[3])
        _model.set_value(row, self.tvwTreeView.position["function_5"], function[4])

    def _do_set_hardware_attributes(self, tree: treelib.Tree) -> None:
        """Set the attributes when the hardware tree is retrieved.

        :param tree: the hardware treelib.Tree().
        :return: None
        :rtype: None
        """
        for _node in tree.all_nodes()[1:]:
            _hardware = _node.data["hardware"]
            self._dic_hardware_attrs[_hardware.hardware_id] = [
                _hardware.name,
                _hardware.part,
            ]

    def _do_set_reliability_attributes(self, tree: treelib.Tree) -> None:
        """Set the attributes when the reliability tree is retrieved.

        :param tree: the reliability treelib.Tree().
        :return: None
        :rtype: None
        """
        for _node in tree.all_nodes()[1:]:
            _reliability = _node.data["reliability"]
            self._dic_reliability_attrs[_reliability.hardware_id] = [
                _reliability.hazard_rate_active,
            ]

    def _on_method_changed(self, method_id: int) -> None:
        """Set method ID attributes when user changes the selection.

        :param method_id: the newly selected allocation method.
        :return: None
        :rtype: None
        """
        self._method_id = method_id
        self._do_set_columns_visible()

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Respond to user changing selected row in Similar Item worksheet.

        :param selection: the Gtk.TreeSelection() that is the newly selected row.
        :return: None
        """
        _attributes = super().on_row_change(selection)

        if _attributes:
            self._record_id = _attributes["hardware_id"]

            pub.sendMessage(
                "selected_similar_item",
                attributes=_attributes,
            )
            pub.sendMessage(
                "request_get_similar_item_attributes",
                node_id=self._record_id,
            )

    def _do_set_columns_visible(self) -> None:
        """Set editable columns based on the Allocation method selected.

        :return: None
        :rtype: None
        """
        if self._method_id == -1:
            self._method_id = 0
        self.tvwTreeView.visible = self._dic_visible_mask[self._method_id]
        self.tvwTreeView.do_set_visible_columns()

    def __do_load_similar_item(
        self, node: Any = "", row: Gtk.TreeIter = None
    ) -> Gtk.TreeIter:
        """Load the similar item RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the hardware tree.
        :return: _new_row; the row that was just populated with hardware data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        _entity = node.data["similar_item"]

        if (
            not self._dic_hardware_attrs[_entity.hardware_id][1]
            and not _entity.parent_id == 0
        ):
            _model = self.tvwTreeView.get_model()

            try:
                _name = self._dic_hardware_attrs[_entity.hardware_id][0]
                _hr_active = self._dic_reliability_attrs[_entity.hardware_id][0]
            except KeyError:
                _name = ""
                _hr_active = 0.0

            _attributes = [
                _entity.revision_id,
                _entity.hardware_id,
                _name,
                _hr_active,
                self._dic_quality[_entity.quality_from_id],
                self._dic_quality[_entity.quality_to_id],
                self._dic_environment[_entity.environment_from_id],
                self._dic_environment[_entity.environment_to_id],
                _entity.temperature_from,
                _entity.temperature_to,
                _entity.change_description_1,
                _entity.change_factor_1,
                _entity.change_description_2,
                _entity.change_factor_2,
                _entity.change_description_3,
                _entity.change_factor_3,
                _entity.change_description_4,
                _entity.change_factor_4,
                _entity.change_description_5,
                _entity.change_factor_5,
                _entity.change_description_6,
                _entity.change_factor_6,
                _entity.change_description_7,
                _entity.change_factor_7,
                _entity.change_description_8,
                _entity.change_factor_8,
                _entity.change_description_9,
                _entity.change_factor_9,
                _entity.change_description_10,
                _entity.change_factor_10,
                _entity.function_1,
                _entity.function_2,
                _entity.function_3,
                _entity.function_4,
                _entity.function_5,
                _entity.result_1,
                _entity.result_2,
                _entity.result_3,
                _entity.result_4,
                _entity.result_5,
                _entity.user_blob_1,
                _entity.user_blob_2,
                _entity.user_blob_3,
                _entity.user_blob_4,
                _entity.user_blob_5,
                _entity.user_float_1,
                _entity.user_float_2,
                _entity.user_float_3,
                _entity.user_float_4,
                _entity.user_float_5,
                _entity.user_int_1,
                _entity.user_int_2,
                _entity.user_int_3,
                _entity.user_int_4,
                _entity.user_int_5,
            ]

            try:
                _new_row = _model.append(row, _attributes)
            except (AttributeError, TypeError, ValueError):
                _new_row = None
                _message = _(
                    f"An error occurred when loading similar item record "
                    f"{node.identifier} into the similar item list.  This might "
                    f"indicate it was missing it's data package, some of the data in "
                    f"the package was missing, or some of the data was the wrong "
                    f"type.  Row data was: {_attributes}"
                )
                pub.sendMessage(
                    "do_log_warning_msg",
                    logger_name="WARNING",
                    message=_message,
                )

        return _new_row



# -*- coding: utf-8 -*-
#
#       ramstk.exim.export.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK Export module."""

# Standard Library Imports
import os
from typing import Any, Dict

# Third Party Imports
# noinspection PyPackageRequirements
import pandas as pd

# noinspection PyPackageRequirements
from openpyxl import load_workbook
from pubsub import pub
from treelib import Tree


class Export:
    """Contains the methods for exporting data from a program database."""

    def __init__(self) -> None:
        """Initialize an Export module instance."""
        # Initialize private dictionary attributes.
        self._dic_output_data: Dict[str, Dict[int, Dict[Any, Any]]] = {}

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._df_output_data: pd.DataFrame = pd.DataFrame()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_load_data, "succeed_get_function_tree")
        pub.subscribe(self._do_load_data, "succeed_get_requirement_tree")
        pub.subscribe(self._do_load_data, "succeed_get_hardware_tree")
        pub.subscribe(self._do_load_data, "succeed_get_validation_tree")
        pub.subscribe(self._do_export, "request_export_data")

    def _do_export(self, file_type: str, file_name: str) -> None:
        """Export selected RAMSTK module data to external file.

        :param file_type: the type of file to export the data to.
            Supported files types are:
                - CSV (using a semi-colon (;) delimiter)
                - Excel
                - Text (using a blank space delimiter)
        :param file_name: the name, with full path, of the file to export
            the RAMSTK Program database data to.
        :return: None
        :rtype: None
        """
        if file_type == "csv":
            self._do_export_to_delimited_text(file_name, separator=";")
        elif file_type == "excel":
            self._do_export_to_excel(file_name)
        elif file_type == "text":
            self._do_export_to_delimited_text(file_name, separator=" ")

    def _do_export_to_delimited_text(self, file_name: str, separator: str):
        """Export RAMSTK project data to a delimited text file.

        :param file_name: the name of the file to export data.
        :param separator: the field delimiter to use.
        :return: None
        :rtype: None
        """
        for (
            __,  # pylint: disable=unused-variable
            _data,
        ) in self._dic_output_data.items():
            self._df_output_data = pd.DataFrame(_data)

            self._df_output_data.to_csv(file_name, sep=separator, index=True)

    # pylint: disable=abstract-class-instantiated
    def _do_export_to_excel(self, file_name: str) -> None:
        """Export RAMSTK project data to an Excel file.

        :param file_name: the name of the file to export data.
        :return: None
        :rtype: None
        """
        _file, _extension = os.path.splitext(file_name)

        for (
            _key,
            _data,
        ) in self._dic_output_data.items():
            self._df_output_data = pd.DataFrame(_data)
            if _extension == ".xls":
                # xlwt can't write each module to a separate sheet so we'll
                # have to make a separate workbook for each work stream module.
                _writer = pd.ExcelWriter(f"{_file}_{_key}.xls", engine="xlwt")
            elif _extension in [".xlsx", ".xlsm"]:
                _writer = pd.ExcelWriter(file_name, engine="openpyxl")
                # Set the writer workbook if it already exists, otherwise
                # just continue.  This allows each work stream module to be
                # written to it's own worksheet.  If the workbook doesn't
                # exist it will be created when the first module is written.
                try:
                    _workbook = load_workbook(file_name)
                    _writer.book = _workbook
                except FileNotFoundError:
                    pass
            else:
                file_name = _file + ".xls"
                _writer = pd.ExcelWriter(file_name, engine="xlwt")
            self._df_output_data.to_excel(_writer, f"{_key}", index=True)
            _writer.save()

            _writer.close()

    def _do_load_data(self, tree: Tree) -> None:
        """Load the attribute data into a Pandas DataFrame.

        :param tree: the data manager tree for the module to export.
        :return: None
        :rtype: None
        """
        _module = tree.get_node(0).tag.lower()
        self._dic_output_data[_module] = {}

        # pylint: disable=unused-variable
        for _node in tree.all_nodes()[1:]:
            _tag = _node.tag
            try:
                self._dic_output_data[_module][_node.identifier] = _node.data[
                    _tag
                ].get_attributes()
            except (KeyError, TypeError):
                pass



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.db.__init__.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK database package."""

# Third Party Imports
from sqlalchemy.ext.declarative import declarative_base

# RAMSTK Local Imports
from .base import BaseDatabase, do_create_program_db

RAMSTK_BASE = declarative_base()



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.hardware.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Hardware Panels."""

# Standard Library Imports
from datetime import date
from typing import Any, Dict, List, Tuple

# Third Party Imports
import treelib
from pubsub import pub

# noinspection PyPackageRequirements
from sortedcontainers import SortedDict

# RAMSTK Package Imports
from ramstk.views.gtk3 import GdkPixbuf, Gtk, _
from ramstk.views.gtk3.widgets import (
    RAMSTKCheckButton,
    RAMSTKComboBox,
    RAMSTKEntry,
    RAMSTKFixedPanel,
    RAMSTKTextView,
    RAMSTKTreePanel,
)


class HardwareTreePanel(RAMSTKTreePanel):
    """Panel to display hierarchy of hardware."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "succeed_retrieve_hardwares"
    _tag = "hardware"
    _title = _("Hardware BoM")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Hardware panel."""
        super().__init__()

        # Initialize private dictionary class attributes.
        self.tvwTreeView.dic_row_loader = {
            "hardware": self.__do_load_hardware,
        }

        # Initialize private list class attributes.

        # Initialize private scalar class attributes.

        # Initialize public dictionary class attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "hardware_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Hardware ID"),
                "gint",
            ],
            "alt_part_number": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Alt. Part Num."),
                "gchararray",
            ],
            "cage_code": [
                3,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("CAGE Code"),
                "gchararray",
            ],
            "comp_ref_des": [
                4,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Comp. Ref. Des."),
                "gchararray",
            ],
            "cost": [
                5,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Cost"),
                "gfloat",
            ],
            "cost_failure": [
                6,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Cost/Failure"),
                "gfloat",
            ],
            "cost_hour": [
                7,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Cost/Hour"),
                "gfloat",
            ],
            "description": [
                8,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Description"),
                "gchararray",
            ],
            "duty_cycle": [
                9,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Duty Cycle"),
                "gfloat",
            ],
            "figure_number": [
                10,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Figure Number"),
                "gchararray",
            ],
            "lcn": [
                11,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("LCN"),
                "gchararray",
            ],
            "level": [
                12,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Level"),
                "gint",
            ],
            "manufacturer_id": [
                13,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Manufacturer"),
                "gint",
            ],
            "mission_time": [
                14,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mission Time"),
                "gfloat",
            ],
            "name": [
                15,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Name"),
                "gchararray",
            ],
            "nsn": [
                16,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("NSN"),
                "gchararray",
            ],
            "page_number": [
                17,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Page Number"),
                "gchararray",
            ],
            "parent_id": [
                18,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Parent ID"),
                "gint",
            ],
            "part": [
                19,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Part?"),
                "gint",
            ],
            "part_number": [
                20,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Part Number"),
                "gchararray",
            ],
            "quantity": [
                21,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Quantity"),
                "gint",
            ],
            "ref_des": [
                22,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Ref. Des."),
                "gchararray",
            ],
            "remarks": [
                23,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Remarks"),
                "gchararray",
            ],
            "repairable": [
                24,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                f"mvw_editing_{self._tag}",
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Repairable?"),
                "gint",
            ],
            "specification_number": [
                25,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Specification"),
                "gchararray",
            ],
            "tagged_part": [
                26,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Tagged Part"),
                "gint",
            ],
            "total_part_count": [
                27,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Part Count"),
                "gint",
            ],
            "total_power_dissipation": [
                28,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Power Dissipation"),
                "gfloat",
            ],
            "year_of_manufacture": [
                29,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                date.today(),
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Year of Manufacture"),
                "gint",
            ],
            "cost_type_id": [
                30,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Cost Type"),
                "gint",
            ],
            "attachments": [
                31,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Attachments"),
                "gchararray",
            ],
            "category_id": [
                32,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Category"),
                "gint",
            ],
            "subcategory_id": [
                33,
                Gtk.CellRendererText(),
                "edited",
                None,
                f"mvw_editing_{self._tag}",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Subcategory"),
                "gint",
            ],
        }
        self.dic_icons = {"assembly": None, "part": None}

        # Initialize public list class attributes.

        # Initialize public scalar class attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _("Displays the hierarchical list of hardware.")
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._on_module_switch,
            "mvwSwitchedPage",
        )

    def _on_module_switch(self, module: str = "") -> None:
        """Respond to changes in selected Module View module (tab).

        :param module: the name of the module that was just selected.
        :return: None
        """
        _model, _row = self.tvwTreeView.selection.get_selected()

        if module == self._tag and _row is not None:
            _comprefdes = _model.get_value(
                _row, self.tvwTreeView.position["comp_ref_des"]
            )
            _name = _model.get_value(_row, self.tvwTreeView.position["name"])
            _title = _(f"Analyzing Hardware item {_comprefdes}: {_name}")

            pub.sendMessage("request_set_title", title=_title)

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Handle events for the Hardware package Module View RAMSTKTreeView().

        This method is called whenever a Hardware Module View RAMSTKTreeView()
        row is activated/changed.

        :param selection: the Hardware class Gtk.TreeSelection().
        :return: None
        """
        _attributes = super().on_row_change(selection)

        if _attributes:
            self._record_id = _attributes["hardware_id"]
            self._parent_id = _attributes["parent_id"]

            _title = _("Analyzing hardware item {0}: {1}").format(
                str(_attributes["comp_ref_des"]), str(_attributes["name"])
            )

            pub.sendMessage(
                "selected_hardware",
                attributes=_attributes,
            )
            pub.sendMessage(
                "request_set_title",
                title=_title,
            )
            for _table in [
                "design_electric",
                "hardware",
                "milhdbk217f",
                "reliability",
            ]:
                pub.sendMessage(
                    f"request_get_{_table}_attributes",
                    node_id=self._record_id,
                )

    def __do_load_hardware(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a hardware item into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the hardware tree.
        :return: _new_row; the row that was just populated with hardware data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        # pylint: disable=unused-variable
        _entity = node.data["hardware"]

        _model = self.tvwTreeView.get_model()

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["assembly"], 22, 22
        )

        if _entity.part == 1:
            # noinspection PyArgumentList
            _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
                self.dic_icons["part"], 22, 22
            )

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.alt_part_number,
            _entity.cage_code,
            _entity.comp_ref_des,
            _entity.cost,
            _entity.cost_failure,
            _entity.cost_hour,
            _entity.description,
            _entity.duty_cycle,
            _entity.figure_number,
            _entity.lcn,
            _entity.level,
            _entity.manufacturer_id,
            _entity.mission_time,
            _entity.name,
            _entity.nsn,
            _entity.page_number,
            _entity.parent_id,
            _entity.part,
            _entity.part_number,
            _entity.quantity,
            _entity.ref_des,
            _entity.remarks,
            _entity.repairable,
            _entity.specification_number,
            _entity.tagged_part,
            _entity.total_part_count,
            _entity.total_power_dissipation,
            _entity.year_of_manufacture,
            _entity.cost_type_id,
            _entity.attachments,
            _entity.category_id,
            _entity.subcategory_id,
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading hardware item {0} into the "
                "hardware tree.  This might indicate it was missing it's data "
                "package, some of the data in the package was missing, or "
                "some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg",
                logger_name="WARNING",
                message=_message,
            )

        return _new_row


class HardwareGeneralDataPanel(RAMSTKFixedPanel):
    """Panel to display general data about the selected Hardware item."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "selected_hardware"
    _tag = "hardware"
    _title = _("Hardware General Information")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Hardware General Date panel."""
        super().__init__()

        # Initialize widgets.
        self.chkRepairable: RAMSTKCheckButton = RAMSTKCheckButton(label=_("Repairable"))
        self.cmbCategory: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbSubcategory: RAMSTKComboBox = RAMSTKComboBox()
        self.txtAltPartNum: RAMSTKEntry = RAMSTKEntry()
        self.txtCompRefDes: RAMSTKEntry = RAMSTKEntry()
        self.txtDescription: RAMSTKTextView = RAMSTKTextView(Gtk.TextBuffer())
        self.txtFigureNumber: RAMSTKEntry = RAMSTKEntry()
        self.txtLCN: RAMSTKEntry = RAMSTKEntry()
        self.txtName: RAMSTKEntry = RAMSTKEntry()
        self.txtPageNumber: RAMSTKEntry = RAMSTKEntry()
        self.txtPartNumber: RAMSTKEntry = RAMSTKEntry()
        self.txtRefDes: RAMSTKEntry = RAMSTKEntry()
        self.txtSpecification: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dictionary instance attributes.
        self.dic_attribute_widget_map = {
            "ref_des": [
                22,
                self.txtRefDes,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The reference designator of the selected hardware item."
                    ),
                },
                _("Reference Designator:"),
                "gchararray",
            ],
            "comp_ref_des": [
                4,
                self.txtCompRefDes,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The composite reference designator of the selected hardware "
                        "item."
                    )
                },
                _("Composite Ref. Des."),
                "gchararray",
            ],
            "name": [
                15,
                self.txtName,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "width": 600,
                    "tooltip": _("The name of the selected hardware item."),
                },
                _("Name:"),
                "gchararray",
            ],
            "description": [
                8,
                self.txtDescription,
                "changed",
                super().on_changed_textview,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "width": 600,
                    "tooltip": _("The description of the selected hardware item."),
                },
                _("Description:"),
                "gchararray",
            ],
            "part_number": [
                20,
                self.txtPartNumber,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _("The part number of the selected hardware item."),
                },
                _("Part Number:"),
                "gchararray",
            ],
            "alt_part_number": [
                2,
                self.txtAltPartNum,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The alternate part number (if any) of the selected hardware "
                        "item."
                    )
                },
                _("Alternate Part Number:"),
                "gchararray",
            ],
            "category_id": [
                32,
                self.cmbCategory,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {},
                _("Category:"),
                "gint",
            ],
            "subcategory_id": [
                33,
                self.cmbSubcategory,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {},
                _("Subcategory:"),
                "gint",
            ],
            "specification_number": [
                25,
                self.txtSpecification,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The specification (if any) governing the selected hardware "
                        "item."
                    ),
                },
                _("Specification:"),
                "gchararray",
            ],
            "page_number": [
                17,
                self.txtPageNumber,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The page number in the governing specification for the "
                        "selected hardware item."
                    )
                },
                _("Page Number:"),
                "gchararray",
            ],
            "figure_number": [
                10,
                self.txtFigureNumber,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The figure number in the governing specification for the "
                        "selected hardware item."
                    )
                },
                _("Figure Number:"),
                "gchararray",
            ],
            "lcn": [
                11,
                self.txtLCN,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The Logistics Control Number (LCN) of the selected hardware "
                        "item."
                    )
                },
                _("LCN:"),
                "gchararray",
            ],
            "repairable": [
                24,
                self.chkRepairable,
                "toggled",
                super().on_toggled,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "Indicates whether or not the selected hardware item is "
                        "repairable."
                    )
                },
                "",
                "gint",
            ],
        }
        self.dicSubcategories: Dict[int, Dict[int, str]] = {}

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.cmbCategory.connect("changed", self._request_load_subcategories)
        self.cmbSubcategory.connect("changed", self._request_load_component)

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_load_subcategories,
            "changed_category",
        )
        pub.subscribe(
            self._do_set_comp_ref_des,
            "succeed_make_comp_ref_des",
        )

    def do_load_categories(self, category: Dict[int, str]) -> None:
        """Load the category RAMSTKComboBox().

        :param category: the dictionary of hardware categories to load.
        :return: None
        :rtype: None
        """
        _model = self.cmbCategory.get_model()
        _model.clear()

        _categories = []
        # pylint: disable=unused-variable
        for __, _key in enumerate(category):
            _categories.append([category[_key]])
        self.cmbCategory.do_load_combo(entries=_categories)  # type: ignore

    def _do_load_subcategories(self, category_id: int) -> None:
        """Load the subcategory RAMSTKComboBox().

        :param category_id: the ID of the selected category.
        :return: None
        """
        _model = self.cmbSubcategory.get_model()
        _model.clear()

        if category_id > 0:
            _subcategories = SortedDict(self.dicSubcategories[category_id])
            _subcategory = []
            for _key in _subcategories:
                _subcategory.append([_subcategories[_key]])
            self.cmbSubcategory.do_load_combo(entries=_subcategory, signal="changed")

    def _do_set_comp_ref_des(self, comp_ref_des: str) -> None:
        """Set the value in the composite reference designator RAMSTKEntry().

        :param comp_ref_des: the composite reference designator value.
        :return: None
        :rtype: None
        """
        self.txtCompRefDes.do_update(
            comp_ref_des,
            signal="changed",
        )

    def _request_load_component(self, combo: RAMSTKComboBox) -> None:
        """Request to load the component widgets.

        :param combo: the RAMSTKComboBox() that called this method.
        :return: None
        """
        pub.sendMessage("changed_subcategory", subcategory_id=combo.get_active())
        for _table in [
            "design_electric",
            "hardware",
            "reliability",
        ]:
            pub.sendMessage(
                f"request_get_{_table}_attributes",
                node_id=self._record_id,
            )

    def _request_load_subcategories(self, combo: RAMSTKComboBox) -> None:
        """Request to have the subcategory RAMSTKComboBox() loaded.

        :param combo: the RAMSTKComboBox() that called this method.
        :return: None
        """
        self._do_load_subcategories(category_id=combo.get_active())
        pub.sendMessage(
            "hardware_category_changed",
            attributes={
                "category_id": combo.get_active(),
            },
        )


class HardwareLogisticsPanel(RAMSTKFixedPanel):
    """Panel to display general data about the selected Hardware task."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "selected_hardware"
    _tag = "hardware"
    _title = _("Hardware Logistics Information")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Hardware Task Description panel."""
        super().__init__()

        # Initialize widgets.
        self.cmbCostType: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbManufacturer: RAMSTKComboBox = RAMSTKComboBox(simple=False)
        self.txtCAGECode: RAMSTKEntry = RAMSTKEntry()
        self.txtCost: RAMSTKEntry = RAMSTKEntry()
        self.txtNSN: RAMSTKEntry = RAMSTKEntry()
        self.txtQuantity: RAMSTKEntry = RAMSTKEntry()
        self.txtYearMade: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "manufacturer_id": [
                13,
                self.cmbManufacturer,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {},
                _("Manufacturer:"),
                "gint",
            ],
            "cage_code": [
                3,
                self.txtCAGECode,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The Commercial and Government Entity (CAGE) Code of the "
                        "selected hardware item."
                    ),
                },
                _("CAGE Code:"),
                "gchararray",
            ],
            "nsn": [
                16,
                self.txtNSN,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                "",
                {
                    "tooltip": _(
                        "The National Stock Number (NSN) of the selected hardware item."
                    )
                },
                _("NSN:"),
                "gchararray",
            ],
            "year_of_manufacture": [
                29,
                self.txtYearMade,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                date.today().year - 2,
                {
                    "width": 100,
                    "tooltip": _(
                        "The year the selected hardware item was introduced to "
                        "the market."
                    ),
                },
                _("Year Introduced:"),
                "gchararray",
            ],
            "quantity": [
                21,
                self.txtQuantity,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                1,
                {
                    "width": 50,
                    "tooltip": _(
                        "The number of the selected hardware items in the design."
                    ),
                },
                _("Quantity:"),
                "gint",
            ],
            "cost": [
                5,
                self.txtCost,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {},
                _("Unit Cost:"),
                "gfloat",
            ],
            "cost_type_id": [
                30,
                self.cmbCostType,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {},
                _("Cost Method:"),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.cmbManufacturer.connect("changed", self._do_load_cage_code)

        # Subscribe to PyPubSub messages.

    def do_load_cost_types(self) -> None:
        """Load the category RAMSTKComboBox().

        :return: None
        """
        self.cmbCostType.do_load_combo([["Assessed"], ["Specified"]])

    def do_load_manufacturers(
        self, manufacturers: Dict[int, Tuple[str, str, str]]
    ) -> None:
        """Load the manufacturer RAMSTKComboBox().

        :param manufacturers: the dictionary with manufacturer information.
            The key is the index from the database table.  The value is a tuple
            with the manufacturer's name, office location, and CAGE code.  An
            example might be:

            ('Sprague', 'New Hampshire', '13606')

        :return: None
        """
        _manufacturer = []
        for _key in manufacturers:
            _manufacturer.append(manufacturers[_key])
        self.cmbManufacturer.do_load_combo(
            entries=_manufacturer,  # type: ignore
            simple=False,
        )

    def _do_load_cage_code(self, combo: RAMSTKComboBox) -> None:
        """Load the CAGE code whenever the manufacturer is changed.

        :param combo: the RAMSTKComboBox() that called this method.
        :return: None
        :rtype: None
        """
        _model = combo.get_model()
        _row = combo.get_active_iter()

        try:
            _cage_code = str(str(_model.get(_row, 2)[0]))
        except TypeError:
            _cage_code = ""

        self.txtCAGECode.do_update(_cage_code, signal="changed")
        pub.sendMessage(
            f"wvw_editing_{self._tag}",
            node_id=[self._record_id, -1],
            package={"cage_code": _cage_code},
        )


class HardwareMiscellaneousPanel(RAMSTKFixedPanel):
    """Panel to display general data about the selected Hardware task."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "selected_hardware"
    _tag = "hardware"
    _title = _("Hardware Miscellaneous Information")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Hardware Task Description panel."""
        super().__init__()

        # Initialize widgets.
        self.chkTagged: RAMSTKCheckButton = RAMSTKCheckButton(label=_("Tagged Part"))
        self.txtAttachments: RAMSTKTextView = RAMSTKTextView(Gtk.TextBuffer())
        self.txtRemarks: RAMSTKTextView = RAMSTKTextView(Gtk.TextBuffer())

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "attachments": [
                31,
                self.txtAttachments,
                "changed",
                super().on_changed_textview,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "height": 150,
                    "width": 600,
                    "tooltip": _(
                        "Hyperlinks to any documents associated with the selected "
                        "hardware item."
                    ),
                },
                _("Attachments:"),
                "gchararray",
            ],
            "remarks": [
                23,
                self.txtRemarks,
                "changed",
                super().on_changed_textview,
                f"mvw_editing_{self._tag}",
                "",
                {
                    "height": 150,
                    "width": 600,
                    "tooltip": _(
                        "Enter any remarks associated with the selected hardware item."
                    ),
                },
                _("Remarks:"),
                "gchararray",
            ],
            "tagged_part": [
                26,
                self.chkTagged,
                "toggled",
                super().on_toggled,
                f"mvw_editing_{self._tag}",
                0,
                {},
                "",
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.hardware.components.relay.py is part of the RAMSTK Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Relay Results Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.milhdbk217f import MilHdbk217FResultPanel
from ramstk.views.gtk3.widgets import RAMSTKEntry


class RelayMilHdbk217FResultPanel(MilHdbk217FResultPanel):
    """Display Relay assessment results attribute data in the RAMSTK Work Book.

    The Relay assessment result view displays all the assessment results
    for the selected relay.  This includes, currently, results for
    MIL-HDBK-217FN2 parts count and MIL-HDBK-217FN2 part stress methods.  The
    attributes of a relay assessment result view are:

    :ivar txtPiC: displays the contact form factor for the relay.
    :ivar txtPiCYC: displays the cycling factor for the relay.
    :ivar txtPiL: displays the load stress factor for the relay.
    :ivar txtPiF: displays the application and construction factor for the
    relay.
    """

    # Define private dict class attributes.
    _dic_part_stress: Dict[int, str] = {
        1: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>L</sub>\u03C0<sub>C</sub>\u03C0<sub"
        ">CYC</sub>\u03C0<sub>F</sub>\u03C0<sub>Q</sub>\u03C0<sub>E</sub"
        "></span>",
        2: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>Q</sub>\u03C0<sub>E</sub></span> ",
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _title: str = _("Relay MIL-HDBK-217F Results")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Relay assessment result view."""
        super().__init__()

        # Initialize widgets
        self.txtPiC: RAMSTKEntry = RAMSTKEntry()
        self.txtPiCYC: RAMSTKEntry = RAMSTKEntry()
        self.txtPiF: RAMSTKEntry = RAMSTKEntry()
        self.txtPiL: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "hazard_rate_model": [
                13,
                self.lblModel,
                "",
                None,
                "",
                "",
                {
                    "tooltip": _(
                        "The assessment model used to calculate the relay hazard rate."
                    ),
                },
                "",
            ],
            "lambda_b": [
                23,
                self.txtLambdaB,
                "",
                None,
                "",
                0.0,
                {
                    "tooltip": _("The base hazard rate for the relay."),
                },
                "\u03BB<sub>b</sub>:",
            ],
            "pi_q": [
                30,
                self.txtPiQ,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The quality factor for the relay."),
                },
                "\u03C0<sub>Q</sub>:",
            ],
            "pi_e": [
                19,
                self.txtPiE,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The environment factor for the relay."),
                },
                "\u03C0<sub>E</sub>:",
            ],
            "pi_c": [
                13,
                self.txtPiC,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The contact form factor for the relay."),
                },
                "\u03C0<sub>C</sub>:",
            ],
            "pi_cyc": [
                18,
                self.txtPiCYC,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The cycling factor for the relay."),
                },
                "\u03C0<sub>CYC</sub>:",
            ],
            "pi_f": [
                20,
                self.txtPiF,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _(
                        "The application and construction factor for the relay."
                    ),
                },
                "\u03C0<sub>F</sub>:",
            ],
            "pi_l": [
                23,
                self.txtPiL,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The load stress factor for the relay."),
                },
                "\u03C0<sub>L</sub>:",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        super().do_set_properties()
        super().do_make_panel()

        # Subscribe to PyPubSub messages.

    def _do_load_entries(self, attributes: Dict[str, Any]) -> None:
        """Load the Relay assessment results widgets.

        :param attributes: the attributes dictionary for the selected relay.
        :return: None
        """
        super().do_load_entries(attributes)

        self.txtPiC.set_sensitive(False)
        self.txtPiCYC.set_sensitive(False)
        self.txtPiF.set_sensitive(False)
        self.txtPiL.set_sensitive(False)

        if self.category_id == 6 and self._hazard_rate_method_id == 2:
            self.txtPiC.do_update(str(self.fmt.format(attributes["piC"])))
            self.txtPiCYC.do_update(str(self.fmt.format(attributes["piCYC"])))
            self.txtPiF.do_update(str(self.fmt.format(attributes["piF"])))
            self.txtPiL.do_update(str(self.fmt.format(attributes["piL"])))



# -*- coding: utf-8 -*-
#
#       ramstk.models.mode.table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Mode Table Model."""

# Standard Library Imports
from collections import defaultdict
from typing import Any, Dict, Type

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.analyses.criticality import (
    calculate_mode_criticality,
    calculate_mode_hazard_rate,
)
from ramstk.models import RAMSTKBaseTable, RAMSTKModeRecord


class RAMSTKModeTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Mode data manager."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_mode_id"
    _db_tablename = "ramstk_mode"
    _select_msg = "selected_revision"
    _tag = "mode"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        """Initialize a Mode data manager instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "hardware_id",
            "mode_id",
        ]

        # Initialize private scalar attributes.
        self._record: Type[RAMSTKModeRecord] = RAMSTKModeRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "mode_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_calculate_criticality, "request_calculate_criticality")

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Any]
    ) -> object:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.hardware_id = attributes["hardware_id"]
        _new_record.mode_id = self.last_id + 1

        return _new_record

    def do_calculate_criticality(self, item_hr: float) -> None:
        """Calculate MIL-STD-1629A, Task 102 criticality of a hardware item.

        :param item_hr: the hazard rate of the hardware item the criticality is
            being calculated for.
        :return: None
        :rtype: None
        """
        _item_criticality: Dict[str, float] = defaultdict(float)
        for _mode in self.tree.children(0):
            _mode.data["mode"].mode_hazard_rate = calculate_mode_hazard_rate(
                item_hr, _mode.data["mode"].mode_ratio
            )
            _mode.data["mode"].mode_criticality = calculate_mode_criticality(
                _mode.data["mode"].mode_hazard_rate,
                _mode.data["mode"].mode_op_time,
                _mode.data["mode"].effect_probability,
            )
            _item_criticality[_mode.data["mode"].severity_class] += _mode.data[
                "mode"
            ].mode_criticality

        pub.sendMessage(
            "succeed_calculate_mode_criticality",
            item_criticality=_item_criticality,
        )



# Standard Library Imports
from typing import Any, Dict, List, Tuple

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk

# RAMSTK Local Imports
from .widget import RAMSTKWidget as RAMSTKWidget

class RAMSTKLabel(Gtk.Label, RAMSTKWidget):
    _default_height: int = ...
    _default_width: int = ...
    def __init__(self, text: str) -> None: ...
    def get_attribute(self, attribute: str) -> Any: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...

def do_make_label_group(
    text: List[str], **kwargs: Dict[str, Any]
) -> Tuple[int, List[RAMSTKLabel]]: ...



# Standard Library Imports
from typing import Any

# Third Party Imports
from treelib import Tree as Tree

class Export:
    _dic_output_data: Any = ...
    _df_output_data: Any = ...
    def __init__(self) -> None: ...
    def _do_export(self, file_type: str, file_name: str) -> None: ...
    def _do_export_to_delimited_text(self, file_name: str, separator: str) -> Any: ...
    def _do_export_to_excel(self, file_name: str) -> None: ...
    def _do_load_data(self, tree: Tree) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.design_electric.components.switch.py is part of the
#       RAMSTK Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Switch Input Panel."""


# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.widgets import RAMSTKComboBox, RAMSTKEntry, RAMSTKFixedPanel


class SwitchDesignElectricInputPanel(RAMSTKFixedPanel):
    """Display Switch assessment input attribute data in the RAMSTK Work Book.

    The Switch assessment input view displays all the assessment inputs for
    the selected switch.  This includes, currently, inputs for
    MIL-HDBK-217FN2.  The attributes of a switch assessment input view are:

    :cvar dict _dic_applications: dictionary of switch applications.  Key is
        switch subcategory ID; values are lists of applications.
    :cvar dict _dic_construction: dictionary of switch construction methods.
        Key is switch subcategory ID; values are lists of construction methods.
    :cvar dict _dic_contact_forms: dictionary of switch contact forms.  Key is
        switch subcategory ID; values are lists of contact forms.

    :ivar cmbApplication: select and display the switch application.
    :ivar cmbConstruction: select and display the switch construction method.
    :ivar cmbContactForm: select and display the switch contact form.
    :ivar txtNCycles: enter and display the number of switch cycles/hour.
    :ivar txtNElements: enter and display the number of switch wafers.
    """

    # Define private dict class attributes.
    # Key is subcategory ID; index is application ID.
    _dic_applications: Dict[int, List[List[str]]] = {
        1: [[_("Resistive")], [_("Inductive")], [_("Lamp")]],
        2: [[_("Resistive")], [_("Inductive")], [_("Lamp")]],
        3: [[_("Resistive")], [_("Inductive")], [_("Lamp")]],
        4: [[_("Resistive")], [_("Inductive")], [_("Lamp")]],
        5: [
            [_("Not Used as a Power On/Off Switch")],
            [_("Also Used as a Power On/Off Switch")],
        ],
    }
    # Key is subcategory ID; index is construction ID.
    _dic_constructions: Dict[int, List[List[str]]] = {
        1: [[_("Snap Action")], [_("Non-Snap Action")]],
        2: [
            [_("Actuation Differential > 0.002 inches")],
            [_("Actuation Differential < 0.002 inches")],
        ],
        3: [[_("Ceramic RF Wafers")], [_("Medium Power Wafers")]],
        5: [[_("Magnetic")], [_("Thermal")], [_("Thermal-Magnetic")]],
    }
    # Key is subcategory ID; index is contact form ID.
    _dic_contact_forms: Dict[int, List[List[str]]] = {
        1: [
            ["SPST"],
            ["DPST"],
            ["SPDT"],
            ["3PST"],
            ["4PST"],
            ["DPDT"],
            ["3PDT"],
            ["4PDT"],
            ["6PDT"],
        ],
        5: [["SPST"], ["DPST"], ["3PST"], ["4PST"]],
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field: str = "hardware_id"
    _select_msg: str = "selected_hardware"
    _tag: str = "design_electric"
    _title: str = _("Switch Design Inputs")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Switch assessment input view."""
        super().__init__()

        # Initialize widgets.
        self.cmbApplication: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbConstruction: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbContactForm: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbQuality: RAMSTKComboBox = RAMSTKComboBox()
        self.txtNCycles: RAMSTKEntry = RAMSTKEntry()
        self.txtNElements: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._quality_id: int = 0

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "quality_id": [
                32,
                self.cmbQuality,
                "changed",
                super().on_changed_combo,
                "wvw_editing_reliability",
                0,
                {
                    "tooltip": _("The quality level of the switch."),
                },
                _("Quality Level:"),
                "gint",
            ],
            "application_id": [
                2,
                self.cmbApplication,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The application of the switch."),
                },
                _("Application:"),
                "gint",
            ],
            "construction_id": [
                6,
                self.cmbConstruction,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The construction method for the switch."),
                },
                _("Construction:"),
                "gint",
            ],
            "contact_form_id": [
                7,
                self.cmbContactForm,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The contact form and quantity of the switch."),
                },
                _("Contact Form:"),
                "gint",
            ],
            "n_cycles": [
                24,
                self.txtNCycles,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _("The number of cycles per hour of the switch."),
                },
                _("Number of Cycles/Hour:"),
                "gfloat",
            ],
            "n_elements": [
                25,
                self.txtNElements,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                1,
                {
                    "tooltip": _("The number of active contacts in the switch."),
                },
                _("Number of Active Contacts:"),
                "gint",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.category_id: int = 0
        self.subcategory_id: int = 0

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_load_comboboxes,
            "changed_subcategory",
        )
        pub.subscribe(
            self._do_set_reliability_attributes,
            "succeed_get_reliability_attributes",
        )

    def do_load_comboboxes(self, subcategory_id: int) -> None:
        """Load the switch RAMSTKComboBox()s.

        :param subcategory_id: the subcategory ID of the selected switch.
        :return: None
        :rtype: None
        """
        self.subcategory_id = subcategory_id

        # Load the quality level RAMSTKComboBox().
        self.cmbQuality.do_load_combo([["MIL-SPEC"], [_("Lower")]], signal="changed")

        # Load the application RAMSTKComboBox().
        try:
            _data = self._dic_applications[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbApplication.do_load_combo(_data, signal="changed")

        # Load the construction RAMSTKComboBox().
        try:
            if self._hazard_rate_method_id == 1:
                _data = [[_("Thermal")], [_("Magnetic")]]
            else:
                _data = self._dic_constructions[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbConstruction.do_load_combo(_data, signal="changed")

        # Load the contact form RAMSTKComboBox().
        try:
            _data = self._dic_contact_forms[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbContactForm.do_load_combo(_data, signal="changed")

    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self._hazard_rate_method_id = attributes["hazard_rate_method_id"]
            self._quality_id = attributes["quality_id"]

            self.cmbQuality.set_sensitive(True)
            self.cmbQuality.do_update(
                self._quality_id,
                signal="changed",
            )

            pub.sendMessage(
                f"request_get_{self._tag}_attributes",
                node_id=self._record_id,
            )

    def _do_set_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set widget sensitivity as needed for the selected switch.

        :return: None
        :rtype: None
        """
        self.cmbApplication.set_sensitive(False)
        self.cmbConstruction.set_sensitive(False)
        self.cmbContactForm.set_sensitive(False)
        self.txtNCycles.set_sensitive(False)
        self.txtNElements.set_sensitive(False)

        if self._hazard_rate_method_id == 1 and self.subcategory_id == 5:
            self.cmbConstruction.set_sensitive(True)
            self.cmbConstruction.do_update(
                attributes["construction_id"],
                signal="changed",
            )
        elif self._hazard_rate_method_id == 2:
            self.cmbApplication.set_sensitive(True)
            self.cmbApplication.do_update(
                attributes["application_id"],
                signal="changed",
            )

            self.__do_set_construction_sensitive(attributes)
            self.__do_set_contact_form_sensitive(attributes)
            self.__do_set_cycles_sensitive(attributes)
            self.__do_set_elements_sensitive(attributes)

    def __do_set_construction_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the construction RAMSTKCombo() sensitive.

        :return: None
        :rtype: None
        """
        if self.subcategory_id in [
            1,
            2,
            3,
            5,
        ]:
            self.cmbConstruction.set_sensitive(True)
            self.cmbConstruction.do_update(
                attributes["construction_id"],
                signal="changed",
            )

    def __do_set_contact_form_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the contact form RAMSTKCombo() sensitive.

        :return: None
        :rtype: None
        """
        if self.subcategory_id in [
            1,
            5,
        ]:
            self.cmbContactForm.set_sensitive(True)
            self.cmbContactForm.do_update(
                attributes["contact_form_id"],
                signal="changed",
            )

    def __do_set_cycles_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the number of cycles RAMSTKEntry() sensitive.

        :return: None
        :rtype: None
        """
        if self.subcategory_id in [
            1,
            2,
            3,
            4,
        ]:
            self.txtNCycles.set_sensitive(True)
            self.txtNCycles.do_update(
                str(attributes["n_cycles"]),
                signal="changed",
            )

    def __do_set_elements_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the number of active elements RAMSTKEntry() sensitive.

        :return: None
        :rtype: None
        """
        if self.subcategory_id in [
            2,
            3,
            4,
        ]:
            self.txtNElements.set_sensitive(True)
            self.txtNElements.do_update(
                str(attributes["n_elements"]),
                signal="changed",
            )



# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord

class RAMSTKFailureDefinitionRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Dict[str, Any]
    __tablename__: str
    __table_args__: Dict[str, bool]
    revision_id: int
    definition_id: int
    definition: str
    def get_attributes(self) -> Dict[str, Any]: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.books.modulebook.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK Module Book Module."""

# Third Party Imports
from pubsub import pub
from treelib import Tree

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk
from ramstk.views.gtk3.function import FunctionModuleView
from ramstk.views.gtk3.hardware import HardwareModuleView
from ramstk.views.gtk3.requirement import RequirementModuleView
from ramstk.views.gtk3.revision import RevisionModuleView
from ramstk.views.gtk3.validation import ValidationModuleView
from ramstk.views.gtk3.widgets import RAMSTKBaseBook


class RAMSTKModuleBook(RAMSTKBaseBook):
    """Display Module Views for the RAMSTK modules.

    Attributes of the Module Book are:

    :ivar dict _dic_module_views: dictionary containing the Module View to
        load into the RAMSTK Module Book for each RAMSTK module.  Key is the
        RAMSTK module name; value is the View associated with that RAMSTK
        module.
    """

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Module Book class.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        RAMSTKBaseBook.__init__(self, configuration)

        # Initialize private dictionary attributes.
        self._dic_module_views = {
            "revision": RevisionModuleView(configuration, logger),
            "function": FunctionModuleView(configuration, logger),
            "requirement": RequirementModuleView(configuration, logger),
            "hardware": HardwareModuleView(configuration, logger),
            "validation": ValidationModuleView(configuration, logger),
        }

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.icoStatus: Gtk.StatusIcon = Gtk.StatusIcon()

        self._set_properties("modulebook")
        self.__make_ui()
        self.__set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_open, "succeed_retrieve_revisions")
        pub.subscribe(self._on_close, "succeed_closed_program")

    def __make_ui(self) -> None:
        """Build the user interface.

        :return: None
        :rtype: None
        """
        self.insert_page(
            self._dic_module_views["revision"],
            tab_label=self._dic_module_views["revision"].hbx_tab_label,
            position=0,
        )

        self.show_all()
        self.set_current_page(0)

    def __set_callbacks(self) -> None:
        """Set callback methods for the RAMSTKListBook and widgets.

        :return: None
        :rtype: None
        """
        self.dic_handler_id["select-page"] = self.connect(
            "select-page", self._on_switch_page
        )
        self.dic_handler_id["switch-page"] = self.connect(
            "switch-page", self._on_switch_page
        )

    def _on_close(self) -> None:
        """Update the Module View when a RAMSTK Program database is closed.

        :return: None
        :rtype: None
        """
        # Remove all the non-Revision pages.
        _n_pages = self.get_n_pages()
        # pylint: disable=unused-variable
        for _page in range(_n_pages - 1):
            self.remove_page(-1)

        # Clear the Revision page treeview.
        _model = self._dic_module_views["revision"].treeview.get_model()
        _model.clear()

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def _on_open(self, tree: Tree) -> None:
        """Insert a page in the module book for each active work stream module.

        :param tree: the work stream module's treelib Tree() containing all
            the data for the work stream module.  Unused in this method,
            but is required as an argument since it is the data package for
            the 'succeed_retrieve_revisions' message.
        :type tree: :class:`treelib.Tree`
        :return: None
        :rtype: None
        """
        for _key in list(self.RAMSTK_USER_CONFIGURATION.RAMSTK_PAGE_NUMBER)[1:]:
            _mkey = self.RAMSTK_USER_CONFIGURATION.RAMSTK_PAGE_NUMBER[_key]
            try:
                _module = self._dic_module_views[_mkey]
                self.insert_page(
                    _module, tab_label=_module.hbx_tab_label, position=_key
                )
            except KeyError:
                pass

        pub.sendMessage("mvwSwitchedPage", module="revision")

    def _on_switch_page(
        self, __notebook: Gtk.Notebook, __page: Gtk.Widget, page_num: int
    ) -> None:
        """Handle page changes in the Module Book Gtk.Notebook().

        :param __notebook: the Tree Book notebook widget.
        :type __notebook: :class:`Gtk.Notebook`
        :param __page: the newly selected page's child widget.
        :type __page: :class:`Gtk.Widget`
        :param page_num: the newly selected page number.
            0 = Revision Tree
            1 = Function Tree
            2 = Requirements Tree
            3 = Hardware Tree
            4 = Software Tree (future)
            5 = Testing Tree (future)
            6 = Validation Tree
            7 = Incident Tree (future)
            8 = Survival Analyses Tree (future)

        :return: None
        :rtype: None
        """
        # Key errors occur when no RAMSTK Program database has been loaded.  In
        # that case, select the Revision page to load.
        # // TODO: Raise error dialog when RAMSTKModuleBook switches page
        # //
        # // The KeyError in RAMSTKModuleBook._on_switch_page() should send
        # // a pubsub message for listeners to respond to.  An error dialog
        # // should be raised for the user to understand why the Revision
        # // module was shown instead.
        try:
            _module = self.RAMSTK_USER_CONFIGURATION.RAMSTK_PAGE_NUMBER[page_num]
        except KeyError:
            _module = "revision"

        pub.sendMessage("mvwSwitchedPage", module=_module)
